class Node {
  static traversal = {
    depth: 0,
    breadth: 1
  };
  static skips = {
    none: 0,
    siblings: 1,
    descendants: 2
  };
  #root = this;
  #parent = null;
  #index = -1;
  #depth = 0;
  #size = 1;
  #children = [];
  get traversal() {
    return Node.traversal;
  }
  get skips() {
    return Node.skips;
  }
  /**
   * The root of the tree to which this node belongs.
   *
   * If the node is a root node, this will be the node itself.
   */
  get root() {
    return this.#root;
  }
  /**
   * The parent of the node.
   *
   * If the node is a root node, this will be `null`.
   */
  get parent() {
    return this.#parent;
  }
  /**
   * The depth of the node in the tree.
   *
   * The depth of a root node is `0`.
   */
  get depth() {
    return this.#depth;
  }
  /**
   * The size of the node.
   *
   * The size of a node is the number of nodes in the subtree rooted at the node.
   */
  get size() {
    return this.#size;
  }
  /**
   * The index of the node in the parent's children.
   *
   * If the node is a root node, this will be `-1`.
   */
  get index() {
    return this.#index;
  }
  /**
   * The next sibling of the node.
   *
   * If the node is the last child of the parent (or does not have a parent), this will be `null`.
   */
  get nextSibling() {
    if (this.#parent === null) {
      return null;
    }
    return this.#parent.#children[this.#index + 1] ?? null;
  }
  /**
   * The previous sibling of the node.
   *
   * If the node is the first child of the parent (or does not have a parent), this will be `null`.
   */
  get previousSibling() {
    if (this.#parent === null) {
      return null;
    }
    return this.#parent.#children[this.#index - 1] ?? null;
  }
  /**
   * The first child of the node.
   *
   * If the node does not have any children, this will be `null`.
   */
  get firstChild() {
    return this.#children[0] ?? null;
  }
  /**
   * The last child of the node.
   *
   * If the node does not have any children, this will be `null`.
   */
  get lastChild() {
    return this.#children[this.#children.length - 1] ?? null;
  }
  /**
   * Checks if the node is a leaf node.
   */
  get isLeaf() {
    return this.#children.length === 0;
  }
  /**
   * Checks if the node is a root node.
   */
  get isRoot() {
    return this.#parent === null;
  }
  /**
   * Checks if the node is the first child of its parent.
   *
   * If the node is a root node, this will be `false`.
   */
  get isFirstChild() {
    return this.#parent !== null && this.#index === 0;
  }
  /**
   * Checks if the node is the last child of its parent.
   *
   * If the node is a root node, this will be `false`.
   */
  get isLastChild() {
    return this.#parent !== null && this.#index === this.#parent.#children.length - 1;
  }
  /**
   * Finds the common ancestor of the node and another node.
   *
   * @param a one of the nodes in question.
   * @param b the other node in question.
   * @returns The common ancestor of the node and the other node, or `null` if there is no common ancestor.
   */
  static commonAncestor(a, b) {
    if (b === a) {
      return b;
    }
    if (b.isAncestorOf(a)) {
      return b;
    }
    if (a.isAncestorOf(b)) {
      return a;
    }
    const deeper = b.#depth > a.#depth ? b : a;
    const shallower = b.#depth > a.#depth ? a : b;
    const potentialAncestors = new Set(shallower.ancestors());
    for (const ancestor of deeper.ancestors()) {
      if (potentialAncestors.has(ancestor)) {
        return ancestor;
      }
    }
    return null;
  }
  /**
   * An iterator that yields the ancestors of the node.
   *
   * The first ancestor is the parent of the node.
   *
   * @returns An iterator that yields the ancestors of the node.
   */
  *ancestors() {
    if (this.#parent === null) {
      return;
    }
    let node = this.#parent;
    while (true) {
      yield node;
      if (node.#parent === null) {
        break;
      }
      node = node.#parent;
    }
  }
  /**
   * An iterator that yields the children of the node.
   *
   * @returns An iterator that yields the children of the node.
   */
  children() {
    return this.#children[Symbol.iterator]();
  }
  /**
   * A generator that yields the descendants of the node.
   *
   * The traversal order of the descendants can be specified.
   * The default traversal order is depth-first.
   * Subtrees or remaining siblings can be skipped when passing Node.skips to the next() method.
   *
   * @param traverse - The traversal order of the descendants.
   * @returns A generator that yields the descendants of the node.
   */
  *descendants(traverse = Node.traversal.depth) {
    const stack = [...this.#children];
    while (stack.length > 0) {
      const node = stack.shift();
      let skips = yield node;
      skips = skips ?? Node.skips.none;
      if (0 !== (Node.skips.siblings & skips) && node.#parent !== null) {
        stack.splice(0, node.#parent.#children.length - node.#index - 1);
      }
      if ((Node.skips.descendants & skips) === 0) {
        if (traverse === Node.traversal.depth) {
          stack.unshift(...node.#children);
        } else {
          stack.push(...node.#children);
        }
      }
    }
  }
  /**
   * An iterator that yields the leaves of the node.
   *
   * @returns An iterator that yields the leaves of the node.
   */
  *leaves() {
    for (const node of this.descendants()) {
      if (node.isLeaf) {
        yield node;
      }
    }
  }
  /**
   * Appends a child to the node.
   *
   * @param child - The child to append.
   * @returns The appended child.
   */
  appendChild(child) {
    if (child.#parent !== null) {
      child.#parent.#removeChild(child);
    }
    this.#children.push(child);
    child.#setParent(this);
    this.#adjustSize(child.#size);
    return child;
  }
  /**
   * Inserts a child before a node.
   *
   * If the node is `null`, the child will be appended to the node.
   *
   * If the child is already a child of the node, it will be moved to the new position.
   *
   * If the child is a child of another node, it will be removed from that node.
   *
   * If the node is not a child of this node, an error will be thrown.
   *
   * @param child - The child to append.
   * @param before - The node to insert the child before.
   * @returns The appended child.
   */
  insertBefore(child, before) {
    if (before === null) {
      return this.appendChild(child);
    }
    if (before.#parent !== this) {
      throw new Error("Provided before node is not a child of this node");
    }
    if (child.#parent !== null) {
      child.#parent.removeChild(child);
    }
    this.#children.splice(before.#index, 0, child);
    child.#setParent(this);
    this.#adjustSize(child.#size);
    for (let i = child.#index + 1; i < this.#children.length; i++) {
      this.#children[i].#index = i;
    }
    return child;
  }
  insertAt(child, index) {
    if (child.#parent !== null) {
      child.#parent.removeChild(child);
    }
    const insertIndex = Math.min(this.#children.length - 1, Math.max(index, 0));
    this.#children.splice(insertIndex, 0, child);
    child.#setParent(this);
    this.#adjustSize(child.#size);
    for (let i = insertIndex + 1; i < this.#children.length; i++) {
      this.#children[i].#index = i;
    }
    return child;
  }
  /**
   * Removes a child from the node.
   *
   * If the child is not a child of the node, an error will be thrown.
   *
   * @param child - The child to remove.
   * @returns The removed child.
   */
  removeChild(child) {
    this.#removeChild(child);
    child.#setParent(null);
    return child;
  }
  /**
   * Detaches the node from the tree.
   *
   * If the node is a root node, this will have no effect.
   *
   * This is the same as calling `removeChild` on the parent of the node.
   */
  detach() {
    if (this.#parent === null) {
      return;
    }
    this.#parent.#removeChild(this);
    this.#setParent(null);
  }
  /**
   * Checks if the node is an ancestor of this node.
   *
   * @param node - The node to check if it is an ancestor of this node.
   * @returns Whether or not the node is an ancestor of this node.
   */
  isAncestorOf(node) {
    for (const ancestor of node.ancestors()) {
      if (ancestor === this) {
        return true;
      }
    }
    return false;
  }
  /**
   * Checks if the node is a descendant of this node.
   *
   * @param node - The node to check if it is a descendant of this node.
   * @returns Whether or not the node is a descendant of this node.
   */
  isDescendantOf(node) {
    return node.isAncestorOf(this);
  }
  /**
   * Checks if the node is a sibling of this node.
   *
   * @param node - The node to check if it is a sibling of this node.
   * @returns Whether or not the node is a sibling of this node.
   */
  isSiblingOf(node) {
    return this.#parent === node.#parent && this !== node;
  }
  #removeChild(child) {
    if (child.#parent !== this) {
      throw new Error("Provided node is not a child of this node");
    }
    this.#adjustSize(-child.#size);
    const index = this.#children.indexOf(child);
    this.#children.splice(index, 1);
    for (let i = index; i < this.#children.length; i++) {
      this.#children[i].#index = i;
    }
  }
  #setParent(parent) {
    this.#parent = parent;
    if (parent === null) {
      this.#root = this;
      this.#index = -1;
      this.#depth = 0;
    } else {
      this.#root = parent.#root;
      this.#index = parent.#children.indexOf(this);
      this.#depth = parent.#depth + 1;
    }
    for (const descendant of this.descendants()) {
      descendant.#root = this.#root;
    }
  }
  #adjustSize(size) {
    this.#size += size;
    for (const ancestor of this.ancestors()) {
      ancestor.#size += size;
    }
  }
}

export { Node };
